/***
	The base class from which a game engine should inherit.

	Version: 1.0
	License: GNU General Public License
*/
module engine;

import SDL;

import std.string;
public import screen;
public import audio;

import std.stdarg;

class Error{}

class Engine{
	const int NoVideo = 0;
	const int Video1024x768 = 1;
	const int Video640x480 = 2;
	const int Video800x600 = 3;
	const int Video320x200 = 4;
	const int Video512x512 = 5;

	enum {up = 0, down, left, right};
	enum {select = 8, start = 9, square = 3, cross = 2, circle = 1, triangle = 0,
		R1 = 7, R2 = 5, L2 = 4, L1 = 6, L3 = 10, R3 = 11};
	typedef int Direction;
	typedef int Buttons;
  public:
	this(int graphics = NoVideo, bool sound = false, int timerClick = 0, bool joystick = false){
		int init = 0;

		if(graphics)
			init |= SDL_INIT_VIDEO;
		if(timerClick)
			init |= SDL_INIT_TIMER;
		if(sound)
			init |= SDL_INIT_AUDIO;
		if(joystick)
			init |= SDL_INIT_JOYSTICK;

		if(SDL_Init(init) == -1 ){
			throw new Error;
		}

		switch(graphics){
			case NoVideo:
				screen = null;
			break;
			case Video1024x768:
				screen = new Screen;
			break;
			case Video640x480:
				screen = new Screen(640, 480);
			break;
			case Video800x600:
				screen = new Screen(800, 600);
			break;
			case Video512x512:
				screen = new Screen(512, 512);
			break;
			case Video320x200:
				screen = new Screen(320, 200);
			break;
			default:
				throw new Error;
		}

		if(timerClick)
			SDL_AddTimer(timerClick, &tcallback, null);

		if(sound)
			audio = new Audio;
		else
			audio = null;

		if(joystick && SDL_NumJoysticks() > 0){
			SDL_JoystickEventState(SDL_ENABLE);
			joyStick = SDL_JoystickOpen(0);
		}
		else
			joyStick = null;


//***********************************************************************

		for(int a = 0; a < 12; a++){
			buttonsDown[a] = 0;
			buttonsChecked[a] = false;
			mapJoystickKeyToButton(a, cast(Buttons) a);
		}
		for(int a = 0; a < 4; a++){
			directionsDown[a] = 0;
			directionsChecked[a] = false;
		}
		for(int a = 0; a < 2; a++){
			stickX[a] = 0;
			stickY[a] = 0;
		}

		wantToQuit = false;


		timers.length = 3;
		calledFunctions.length = 3;
	}

	~this(){
		if(joyStick)
			SDL_JoystickClose(joyStick);
		delete audio;
		delete screen;

		SDL_Quit();
	}

	void run(){
		eventLoop();
	}

	void setTitle(char[] title){
		SDL_WM_SetCaption(std.string.toStringz(title), null);
	}

	bool buttonWasPressed(Buttons button){
		if(!buttonsChecked[button] && buttonsDown[button]){
			buttonsChecked[button] = true;
			return true;
		}
		return false;
	}

	bool directionWasPressed(Direction direction){
		if(!directionsChecked[direction] && directionsDown[direction]){
			directionsChecked[direction] = true;
			return true;
		}
		return false;
	}

	bool buttonIsDown(Buttons button){
		if(button < 12 && button >= 0)
			return buttonsDown[button];
		return false;
	}

	bool directionIsDown(Direction d){
		if(d < 4 && d >= 0)
			return directionsDown[d];
		return false;
	}


	int getStickX(int stick){
		if( stick >= 0 && stick < 2)
			return stickX[stick];
		else
			return 0;
	}

	int getStickY(int stick){
		if( stick >= 0 && stick < 2)
			return stickY[stick];
		else
			return 0;
	}

	Point mouseLocation(){
		return XY(mouseX, mouseY);
	}

	void quit(){
		wantToQuit = true;
	}



  protected:
	const int BUTTONDOWN = 0;
	const int BUTTONUP   = 1;
	const int MOTION     = 2;

	void keyEvent(int type, int key, int character, int modifiers){
		if(character == 'q' || key == 'q')
			quit();
	}

	void mouseEvent(){

	}

	void joystickEvent(int type, int which, int button, int state){

	}

	bool quitEvent(){
		return true;
	}

	void timerEvent(){
		int s = stackLocation;
		if(s < 0 || calledFunctions[s] is null)
			quit();
		else{
			bool result = calledFunctions[s](timers[s]);
			timers[s]++;
			if(!result){
				stackLocation--;
				timerEvent();	// possible hack
			}
		}
	}

	/**
		Returns: true if it needs to continue
			 false if it has finished its task
	*/
	bool delegate(int) calledFunctions[];
	int[] timers;

	int stackLocation;

	void call(bool delegate(int) what){
		stackLocation++;
		if(stackLocation == timers.length){
			timers.length = timers.length + 3;
			calledFunctions.length = calledFunctions.length + 3;
		}

		calledFunctions[stackLocation] = what;
		timers[stackLocation] = 0;
	//	timerEvent();				// possible hack








		SDL_Event event;

		event.type = SDL_USEREVENT;
		event.user.code = 0;
		event.user.data1 = null;
		event.user.data2 = null;
		SDL_PushEvent(&event);





	}

	void jump(bool delegate(int) what){
		call(what);
	//	calledFunctions[stackLocation] = what;
	//	timers[stackLocation] = 0;
	}
//	void branch()

	Screen screen;
	Audio audio;

  private:
	SDL_Joystick* joyStick;
	int[int] keyboardMap;
	int[int] joystickMap;

	bool wantToQuit;

	bool buttonsDown[12];
	bool directionsDown[4];
	int stickX[2];
	int stickY[2];

	bool buttonsChecked[12];
	bool directionsChecked[4];

	int mouseX;
	int mouseY;

	void mapJoystickKeyToButton(int a, Buttons b){
		joystickMap[a] = cast(int) b;
	}

	int eventLoop(){
		SDL_Event event;
	  while(SDL_WaitEvent(&event) >= 0 && !wantToQuit){
	    switch(event.type){
		case SDL_KEYUP:
		case SDL_KEYDOWN:
			bool type = event.key.type == SDL_KEYDOWN ? true : false;
			switch(event.key.keysym.sym){
				case SDLK_UP:
					directionsDown[up] = type;
					directionsChecked[up] = false;
				break;
				case SDLK_DOWN:
					directionsDown[down] = type;
					directionsChecked[down] = false;
				break;
				case SDLK_LEFT:
					directionsDown[left] = type;
					directionsChecked[left] = false;
				break;
				case SDLK_RIGHT:
					directionsDown[right] = type;
					directionsChecked[right] = false;
				break;
				default:
					if(event.key.keysym.sym in keyboardMap){
						int button = keyboardMap[event.key.keysym.sym];
						buttonsDown[button] = type;
						buttonsChecked[button] = false;
					}
			}


			keyEvent(
				event.key.type == SDL_KEYDOWN ? BUTTONDOWN : BUTTONUP,
				event.key.keysym.sym,
				event.key.keysym.unicode,
				event.key.keysym.mod
			);
		break;
		case SDL_JOYAXISMOTION:
		if ( ( event.jaxis.value < -3200 ) || (event.jaxis.value > 3200 ) || event.jaxis.value == 0){
			int stick;
			if(event.jaxis.axis >= 0 && event.jaxis.axis < 6)

				stick = event.jaxis.axis;
			else
				break;
			switch(stick){
			  case 4:
			  case 0:
				directionsDown[left] = event.jaxis.value < -28000;
				directionsDown[right] = event.jaxis.value > 28000;
			/* fall through */
			  case 2:
				if(stick == 0)
					stickX[0] = event.jaxis.value;
				if(stick == 2)
					stickX[1] = event.jaxis.value;
			  break;
			  case 5:
			  case 1:
				directionsDown[up] = event.jaxis.value < -28000;
				directionsDown[down] = event.jaxis.value > 28000;
			/* fall through */
			  case 3:
				if(stick == 1)
					stickY[0] = event.jaxis.value;
				if(stick == 3)
					stickY[1] = event.jaxis.value;
			}

			joystickEvent(
				MOTION,
				event.jaxis.which,
				event.jaxis.axis,
				event.jaxis.value
			);
		}
		break;
		case SDL_JOYHATMOTION:
		//	directionsDown[up] = (event.jaxis.value & SDL_HAT_UP) ? true : false;
		//	directionsDown[down] = (event.jaxis.value & SDL_HAT_DOWN) ? true : false;
		//	directionsDown[left] = (event.jaxis.value & SDL_HAT_LEFT) ? true : false;
		//	directionsDown[right] = (event.jaxis.value & SDL_HAT_RIGHT) ? true : false;
		/+
			joystickEvent(
				HATMOTION,
				event.jhat.which,
				event.jhat.hat,
				event.jhat.value
			);
		+/
		break;
		case SDL_JOYBUTTONDOWN:
		case SDL_JOYBUTTONUP:
			int button = joystickMap[event.jbutton.button];

			buttonsDown[button] = event.jbutton.type == SDL_JOYBUTTONDOWN ? true : false;
			buttonsChecked[button] = false;

			joystickEvent(
				event.jbutton.type == SDL_JOYBUTTONDOWN ? BUTTONDOWN : BUTTONUP,
				event.jbutton.which,
				event.jbutton.button,
				event.jbutton.state
			);
		break;

		case SDL_MOUSEBUTTONDOWN:
		case SDL_MOUSEBUTTONUP:
		/+
			mouseEvent(
				event.button.type == SDL_MOUSEBUTTONDOWN ? BUTTONDOWN : BUTTONUP,
				event.button.x,
				event.button.y,
				0,		// xrel
				0,		// yrel
				event.button.button,
				0		//state
			);
		+/
		break;
		case SDL_MOUSEMOTION:
			mouseX = event.motion.x;
			mouseY = event.motion.y;
		/+
			mouseEvent(
				MOTION,
				event.motion.x,
				event.motion.y,
				event.motion.xrel,
				event.motion.yrel,
				0,
				event.motion.state
			);
		+/
		break;

		case SDL_USEREVENT:
			timerEvent();
		break;
		case SDL_QUIT:
			if(quitEvent() == true)
				quit();
		break;
		default:
	    }
	  }
		return 0;
	}

}
extern(C){
	Uint32 tcallback(Uint32 interval, void* param){
		SDL_Event event;

		event.type = SDL_USEREVENT;
		event.user.code = 0;
		event.user.data1 = null;
		event.user.data2 = null;
		SDL_PushEvent(&event);

		return interval;
	}
}
